<?php

namespace App\Command;

use App\Entity\SentenceListeningNote;
use App\Entity\SentenceNote;
use App\Entity\Term;
use App\Service\AnkiService;
use App\Utils\Japanese;
use App\Utils\Progress;
use Symfony\Component\Console\Attribute\AsCommand;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

#[AsCommand(
    name: 'app:create-production',
    description: 'Add a short description for your command',
)]
class CreateProductionCommand extends Command
{
    public function __construct(
        private AnkiService $ankiService,
    ) {
        parent::__construct();
    }

    private static function extractKanji(string $str): array
    {
        preg_match_all('/\p{Script=Han}/u', $str, $matches);
        return array_unique($matches[0]);
    }

    private static function getOnlyKanji(string $str): string
    {
        return preg_replace('/[^\p{Script=Han}]/u', '', $str);
    }

    private static function kanjiDiff(array &$ref, string $subject): bool
    {
        $len = mb_strlen($subject);
        $hasUnseenKanji = false;

        for ($i = 0; $i < $len; $i++) {
            $subKanji = mb_substr($subject, $i, 1);

            foreach ($ref as $refKanji => $value) {
                if ($subKanji === $refKanji) continue 2;
            }

            $ref[$subKanji] = 0;
            $hasUnseenKanji = true;
        }

        return $hasUnseenKanji;
    }

    protected function configure(): void
    {
        $this
            ->addArgument('count', InputArgument::OPTIONAL, 'Amount of cards to make', 1);
            //->addOption('option1', null, InputOption::VALUE_NONE, 'Option description')
        ;
    }

    protected function createProductionNoteFromTerm(Term $term): void
    {
        $noteIds = $this->ankiService->findNotesIds(sprintf(
            '"SentKanji:*%s*" "note:%s"',
            $term->getKanji(),
            SentenceNote::MODEL_NAME,
        ));

        if (count($noteIds) <= 0) {
            $noteIds = $this->ankiService->findNotesIds(sprintf(
                '"VocabKanji:*%s*" "note:%s"',
                $term->getKanji(),
                SentenceNote::MODEL_NAME,
            ));
        }

        $sNote = $this->ankiService->getNote($noteIds[array_key_last($noteIds)]);

        $newSlNote = SentenceListeningNote::fromNote($sNote, $term);
        if (!$this->ankiService->addNote($newSlNote, 'production')) {
            throw new \Exception('Failed to add note!');
        }
    }

    /** @return list<SentenceNote> */
    private function getAllSentenceNotes(): array
    {
        printf('Getting all SentenceNote...');
        $allIds   = $this->ankiService->getAllSentenceNoteIds();
        $allNotes = $this->ankiService->getNotes($allIds);
        printf(" OK (%d)\n", count($allNotes));
        return $allNotes;
    }

    /** @return list<SentenceListeningNote> */
    private function getAllSentenceListeningNotes(): array
    {
        printf('Getting all SentenceListeningNote...');
        $allListeningIds   = $this->ankiService->getAllSentenceListeningNoteIds();
        $allListeningNotes = $this->ankiService->getNotes($allListeningIds);
        printf(" OK (%d)\n", count($allListeningNotes));
        return $allListeningNotes;
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $allSentenceNotes = $this->getAllSentenceNotes();
        $allListeningNotes = $this->getAllSentenceListeningNotes();

        // --------- Getting list<SentenceNote> into array<TermKanji, Term> ---

        $allTerms = [];
        $knownKanji = [];
        $termCounts = [];
        printf('Indexing all terms...');
        foreach ($allSentenceNotes as $note) {
            foreach ($note->getTerms() as &$term) {
                assert($term instanceof Term);

                if (key_exists($term->getKanji(), $allTerms)) continue;
                $termCounts[$term->getKanji()] = 0;
                $allTerms[$term->getKanji()] = &$term;
                foreach (self::extractKanji($term->getKanji()) as $kanji) {
                    $knownKanji[$kanji] = 0;
                }
            }
        }
        printf(" OK (%d)\n", count($allTerms));


        $progress = new Progress('Getting frequenciees', count($allSentenceNotes));
        foreach ($allSentenceNotes as $note) {
            $progress->tick();

            $sentKanji = str_replace(
                "\u{200E}",
                '',
                strip_tags($note->getFields()['SentKanji'])
            );

            foreach ($knownKanji as $kanji => &$count) {
                if (str_contains($sentKanji, $kanji)) $count++;
            }
        }
        printf("\n");


        $seenKanji = $this->ankiService->getKnownSlnKanjiCounts();
        printf('Rating terms...');
        foreach ($allTerms as $key => $term) {
            $count = Japanese::kanjiDiff(
                $term->getKanji(),
                array_keys($seenKanji),
            );

            if ($count <= 0) {
                unset($allTerms[$key]);
                unset($termCounts[$key]);
                continue;
            }

            $termCounts[$term->getKanji()] = $count;
        }
        uksort($termCounts, function ($a, $b) {
            $aLen = mb_strlen(Japanese::getOnlyKanji($a));
            $bLen = mb_strlen(Japanese::getOnlyKanji($b));
            return $bLen <=> $aLen;
        });
        asort($termCounts);
        printf(" OK\n");

        // Have into account the ones that have already been created.
        // This will not only skip them but also update the general array for
        // already seen kanji.
        $seenKanji = [];
        printf('Filtering out terms with no new kanji...');
        // First pass: Get the list of the kanji we've seen
        foreach ($allListeningNotes as $listeningNote) {
            assert($listeningNote instanceof SentenceListeningNote);

            $termKanji = self::getOnlyKanji($listeningNote->getTerm()->getKanji());
            self::kanjiDiff($seenKanji, $termKanji);
        }
        // Second pass: Remove terms with no new kanji at all
        foreach ($termCounts as $term => $count) {
            $termKanji = self::getOnlyKanji($term);

            if (!self::kanjiDiff($seenKanji, $termKanji)) {
                unset($termCounts[$term]);
                unset($allTerms[$term]);
            }
        }
        printf(" OK\n");

        asort($termCounts);
        printf("\n");


        $newNotesCount = intval($input->getArgument('count'));

        foreach ($termCounts as $term => $count) {
            if ($newNotesCount <= 0) break;

            // FIXME: This shouldn't happen at all
            if (!$allTerms[$term] instanceof Term) continue;

            $termKanji = self::getOnlyKanji($term);
            printf("%s: %d\n", $term, $count);

            //$len = mb_strlen($termKanji);
            //for ($i = 0; $i < $len; $i++) {
            //    $iKanji = mb_substr($termKanji, $i, 1);
            //    printf(" - %s: %0.2f\n", $iKanji, $knownKanji[$iKanji] / $len);
            //}

            $this->createProductionNoteFromTerm($allTerms[$term]);
            $newNotesCount -= 1;
        };

        printf(
            <<<FMNT
            total:         %d cards
            max usage:     %0.2f MiB
            current usage: %0.2f MiB\n
            FMNT,
            count($termCounts),
            memory_get_peak_usage() / 1024 / 1024,
            memory_get_usage() / 1024 / 1024,
        );

        return Command::SUCCESS;
    }
}
